/********************************************************
 * mtstatus -- Report the status of a Exabyte tape 	*
 *							*
 * Usage:						*
 *	ex_status [-a] [-e] [-r] device			*
 *							*
 * 		-a -- Print all reports			*
 * 		-e -- Error report			*
 * 		-r -- Report raw data			*
 *							*
 * $Header: $						*
 ********************************************************/
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#ifdef solaris
#include <sys/scsi/generic/mode.h>
#include <sys/scsi/generic/commands.h>
#include <sys/scsi/impl/types.h>
#include <sys/scsi/impl/uscsi.h>
#include <sys/systeminfo.h>
#endif /* solaris */
#ifdef linux
#include <scsi/scsi.h>
#include <scsi_ioctl.h>
#endif /* linux */

const int EX_LOG_SENSE = 0x4D;	/* Read Log Sense */

static int raw = 0;		/* Output raw data */
static int fd;			/* File descriptor of the tape */

enum page_code {
    P_READ_ERROR  = 0x02,		/* Read error counter */
    P_WRITE_ERROR = 0x03		/* Write error counter */
};

/* The information about each page we're getting */
struct page_info {
   enum page_code code;		/* Page code number */
   char *name;			/* The name of the page */
};

static struct page_info page_list[] = {
	{0,		"Supported pages"},
	{P_READ_ERROR,	"Write Error Counters"},
	{P_WRITE_ERROR,	"Read error Counters"},
	{-1,		NULL}
};

struct log_info {
    unsigned char page_code;		/* Page code from the log */
    unsigned char reserved;		/* Reserved byte */
    unsigned char size_code[2];		/* Size of the data */
};
union scsi_log_struct {
    struct log_info info;	/* Information for a log */
    unsigned char raw[500];	/* Raw SCSI data */
};
/* The log data from the tape drive */
static union scsi_log_struct log;

/********************************************************
 * raw_page -- Print the page in the raw		*
 ********************************************************/
static void raw_page(void)
{
    int i;	/* Index for our dump */
    printf("----------------------------------------------------\n");

    for (i = 0; i < log.info.size_code[1]+4; ++i) {
	printf("%2d: 0x%02x\n", i, log.raw[i]);
    }
    printf("\n\n\n");
}

/* Turn variable into a signed character */
#define SC(what) ((signed char)(what))

/* Turn variable into a unsigned long long */
#define ULL(value, index) \
     ((unsigned long long)((value)[index]))

/* Turn variable into a unsigned long */
#define UL(value, index) \
     ((unsigned long)((value)[index]))

/* Turn variable into a unsigned int */
#define U(value, index) \
     ((unsigned int)((value)[index]))

/* Turn variable into a unsigned int */
#define L(value, index) \
     ((unsigned int)(SC((value)[index])))

/* Turn two bytes into a a short int */
#define TWO_BYTES(value) (signed short int) (\
    ((U(value,0) << 8) | U(value, 1))	    )

/* Turn three bytes into a a short int */
#define THREE_BYTES(value) \
    ((L( value, 0) << 16L) | \
     (UL(value, 1) << 8L)  | \
      UL(value, 2))

/* Turn four bytes into a a short int */
#define FOUR_BYTES(value) \
   ((UL(value, 0) << 24L) | \
    (UL(value, 1) << 16L) | \
    (UL(value, 2) << 8L)  | \
     UL(value, 3))

#define FIVE_BYTES(value) \
   ((ULL(value, 0) << 32L) | \
    (ULL(value, 1) << 24L) | \
    (ULL(value, 2) << 16L) | \
    (ULL(value, 3) << 8L)  | \
     ULL(value, 4))
/********************************************************
 * print_param2 -- print a two byte parameter		*
 *							*
 * Returns						*
 *	Pointer to first character after the data	*
 ********************************************************/
static u_char *print_param2(
    const char name[],		/* Name of the parameter */
    u_char *cur_ptr		/* Pointer to the data */
) {
    printf("%s Parameter Code %d\n", name, TWO_BYTES(cur_ptr));
    cur_ptr += 2;

    printf("%s Flag 0x%02x\n", name, *cur_ptr);
    cur_ptr += 1;

    printf("%s Length %d\n", name, *cur_ptr);
    if (*cur_ptr != 2) printf("*** Length incorrect\n");
    cur_ptr += 1;

    printf("%s value %d\n", name, TWO_BYTES(cur_ptr));
    cur_ptr += 2;

    printf("\n");
    return (cur_ptr);
}
/********************************************************
 * print_param3 -- print a three byte parameter		*
 *							*
 * Returns						*
 *	Pointer to first character after the data	*
 ********************************************************/
static u_char *print_param3(
    const char name[],		/* Name of the parameter */
    u_char *cur_ptr		/* Pointer to the data */
) {
    printf("%s Parameter Code %d\n", name, TWO_BYTES(cur_ptr));
    cur_ptr += 2;

    printf("%s Flag 0x%02x\n", name, *cur_ptr);
    cur_ptr += 1;

    printf("%s Length %d\n", name, *cur_ptr);
    if (*cur_ptr != 3) printf("*** Length incorrect\n");
    cur_ptr += 1;

    printf("%s value %ld\n", name, THREE_BYTES(cur_ptr));
    cur_ptr += 3;

    printf("\n");
    return (cur_ptr);
}
/********************************************************
 * print_param4 -- print a four byte parameter		*
 *							*
 * Returns						*
 *	Pointer to first character after the data	*
 ********************************************************/
static u_char *print_param4(
    const char name[],		/* Name of the parameter */
    u_char *cur_ptr		/* Pointer to the data */
) {
    printf("%s Parameter Code %d\n", name, TWO_BYTES(cur_ptr));
    cur_ptr += 2;

    printf("%s Flag 0x%02x\n", name, *cur_ptr);
    cur_ptr += 1;

    printf("%s Length %d\n", name, *cur_ptr);
    if (*cur_ptr != 4) printf("*** Length incorrect\n");
    cur_ptr += 1;

    printf("%s value %ld\n", name, FOUR_BYTES(cur_ptr));
    cur_ptr += 4;

    printf("\n");
    return (cur_ptr);
}

/********************************************************
 * print_page_head -- print the page information	*
 *							*
 * Returns						*
 *	Pointer to first character after the data	*
 ********************************************************/
static u_char *print_page_head(
    char *cur_ptr	/* Pointer to the current character */
)
{
    printf("Page code 0x%02x\n", *cur_ptr);
    cur_ptr += 1;

    printf("Reserved1 %d\n", *cur_ptr);
    cur_ptr += 1;

    printf("Length %d\n", TWO_BYTES(cur_ptr));
    cur_ptr += 2;
    printf("\n");
    return (cur_ptr);
}
/********************************************************
 * print_read_write -- Print the read/write pages	*
 ********************************************************/
static void print_read_write(
    const char what[]	/* What we're dumping */
) {
    u_char *cur_ptr;	/* Pointer to current character in the data */

    printf("%s Error Counter Pages\n", what);
    cur_ptr = &log.raw[0];

    cur_ptr = print_page_head(cur_ptr);

    cur_ptr = print_param2("Soft Errors", cur_ptr);
    cur_ptr = print_param4(
		 "Total Times Correction Algorithm Processed", cur_ptr);
    cur_ptr = print_param4("Total Groups Processed", cur_ptr);
    cur_ptr = print_param2("Hard Errors ", cur_ptr);
    printf("\n");
}

/********************************************************
 * print_tape_param -- Print a tape log page parameter	*
 ********************************************************/
static u_char *print_tape_param(
    const char name[],	/* Name of the parameter */
    u_char *cur_ptr	/* Pointer to current character in the data */
) {
    char full_name[100];	/* Full name of the page */

    sprintf(full_name, "%s Groups Written", name);
    cur_ptr = print_param3(full_name, cur_ptr);

    sprintf(full_name, "%s Raw Retries", name);
    cur_ptr = print_param2(full_name, cur_ptr);

    sprintf(full_name, "%s Read", name);
    cur_ptr = print_param3(full_name, cur_ptr);

    sprintf(full_name, "%s ECC Retries", name);
    cur_ptr = print_param2(full_name, cur_ptr);

    return (cur_ptr);
}

/********************************************************
 * print_tape_log -- print the tape log page		*
 ********************************************************/
static void print_tape_log(void)
{
    u_char *cur_ptr;	/* Pointer to current character in the data */

    cur_ptr = &log.raw[0];

    printf("Tape Log Page\n");
    cur_ptr = print_page_head(cur_ptr);
    cur_ptr = print_tape_param("Current", cur_ptr);
    cur_ptr = print_tape_param("Previous", cur_ptr);
    cur_ptr = print_param4("Total Written", cur_ptr);
    cur_ptr = print_param3("Total Raw Retries", cur_ptr);
    cur_ptr = print_param4("Total Read", cur_ptr);
    cur_ptr = print_param3("Total Ecc", cur_ptr);
}
/********************************************************
 * print_mechanism -- Print mechanism page		*
 ********************************************************/
static void print_mechanism(void)
{
    u_char *cur_ptr;	/* Pointer to the current data */

    cur_ptr = &log.raw[0];

    printf("Tape Mechanism Counters Page\n");
    cur_ptr = print_page_head(cur_ptr);

    cur_ptr = print_param2("Faults 12v", cur_ptr);
    cur_ptr = print_param2("Mode Sensor Fault", cur_ptr);
    cur_ptr = print_param2("Tension Too Low", cur_ptr);
    cur_ptr = print_param2("Bad Diameter", cur_ptr);
    cur_ptr = print_param2("Capstan Stalled", cur_ptr);
    cur_ptr = print_param2("DPG Calibration Failed", cur_ptr);
    cur_ptr = print_param2("Drum Stalled", cur_ptr);
    cur_ptr = print_param2("Drum has Lost Lock", cur_ptr);
    cur_ptr = print_param2("Drum PG Lost", cur_ptr);
    cur_ptr = print_param2("###", cur_ptr);
}
/********************************************************
 * print_compression -- Print compression page		*
 ********************************************************/
static void print_compression(void)
{
    u_char *cur_ptr;	/* Pointer to the current data */

    cur_ptr = &log.raw[0];

    printf("Tape Compression Transfer Page\n");
    cur_ptr = print_page_head(cur_ptr);

    cur_ptr = print_param4("Number of entries written", cur_ptr);
    cur_ptr = print_param4("Number of entries read", cur_ptr);
    cur_ptr = print_param4("Number of records written", cur_ptr);
    cur_ptr = print_param4("Number of records read", cur_ptr);
    cur_ptr = print_param4("Kilobytes to data compression  ", cur_ptr);
    cur_ptr = print_param4("Kilobytes from data compression", cur_ptr);
    cur_ptr = print_param4("Kilobytes to tape  ", cur_ptr);
    cur_ptr = print_param4("Kilobytes from tape", cur_ptr);
    cur_ptr = print_param4("Logical entity size", cur_ptr);
    cur_ptr = print_param4("Physical entity size", cur_ptr);
    cur_ptr = print_param4("Uncompressed entities", cur_ptr);
}


/********************************************************
 * print_page -- print the current log page		*
 ********************************************************/
static void print_page(void)
{
    switch (log.raw[0]) {
	case 0x02:
	    if (raw) raw_page();
	    print_read_write("Read");
	    break;
	case 0x03:
	    if (raw) raw_page();
	    print_read_write("Write");
	    break;
	case 0x30:
	    if (raw) raw_page();
	    print_tape_log();
	    break;
	case 0x38:
	    if (raw) raw_page();
	    print_mechanism();
	    break;
	case 0x39:
	    if (raw) raw_page();
	    print_compression();
	    break;
	default:
	    raw_page();
	    break;
    }
}
/********************************************************
 * get_page -- Get a page of the LOG_SENSE data		*
 *							*
 * Data is returned in the global structure log		*
 ********************************************************/
static void get_page(
    const int page	/* Page index */
)
{
#ifdef linux
    static struct log_cmd {
	unsigned int inlen;		 /* Length of data going in */
	unsigned int outlen;		 /* Length of data going out */
	unsigned char cmd[10];		 /* The command to execute */
	unsigned char data[sizeof(log)]; /* The data buffer */
    } scsi_cmd;

    scsi_cmd.cmd[0] = EX_LOG_SENSE;
    scsi_cmd.cmd[1] = 0;
    scsi_cmd.cmd[2] = page;

    scsi_cmd.cmd[3] = 0;
    scsi_cmd.cmd[4] = 0;
    scsi_cmd.cmd[5] = 0;
    scsi_cmd.cmd[6] = 0;

    scsi_cmd.cmd[7] = sizeof(log)/256;
    scsi_cmd.cmd[8] = sizeof(log)%256;
    scsi_cmd.cmd[9] = 0;

    scsi_cmd.inlen = 0;
    scsi_cmd.outlen = sizeof(log);

    if (ioctl (fd, SCSI_IOCTL_SEND_COMMAND, &scsi_cmd) < 0) { 
	int error = errno;
	perror("SCSI command failed: ");
	printf("Error code: %d\n", error);
	exit (1);
    }

    /* Copy the the data to the log variable */
    /* (I know it's a hack) */
    memcpy(&log, scsi_cmd.cmd, sizeof(log));

#endif /* linux */
   
#ifdef solaris
    struct uscsi_cmd cmd;	/* The scsi command */
    union scsi_cdb cdb;	/* The command/data block */

    memset(&cmd, '\0', sizeof(cmd));
    memset(&cdb, '\0', sizeof(cdb));

    cdb.scc_cmd = HP_LOG_SENSE;
    cdb.scc_lun = 0;	
    cdb.g1_count1 = sizeof(log) / 256;
    cdb.g1_count0 = sizeof(log) % 256;
    cdb.g1_addr3  = page;

    cmd.uscsi_cdb=(caddr_t)&cdb;
    cmd.uscsi_cdblen=10;
    cmd.uscsi_bufaddr=(caddr_t)&log;
    cmd.uscsi_buflen=sizeof(log);
    cmd.uscsi_flags=USCSI_DIAGNOSE|USCSI_SILENT|USCSI_ISOLATE|USCSI_READ;

    if (ioctl (fd, USCSICMD, &cmd) < 0) { 
	int error = errno;
	perror("SCSI command failed: ");
	printf("Error code: %d\n", error);
	exit (1);
    }
#endif /* solaris */
}
/********************************************************
 * report_all -- report all data pages.			*
 ********************************************************/
static void report_all(void)
{
    int page;	/* Current page number */

    for (page = 0; page_list[page].code != -1; ++page) {
	printf("Page: 0x%02x %s\n", 
		page_list[page].code, page_list[page].name);

	get_page(page_list[page].code);
	print_page();
    }
}

/********************************************************
 * report_errors -- print a error counter report	*
 *							*
 * Page format						*
 * 	0,1 -- page code				*
 * 	2 -- Flags					*
 *	3 -- page length (1c)				*
 *							*
 *	4,5 -- 0,1 param code (3)			*
 *	6 -- 2 flags					*
 *	7 -- 3 param length (2)				*
 *	8,9 -- 4,5 value				*
 *							*
 *	10,11 -- 0,1 param code (4)			*
 *	12 -- 2 flags					*
 *	13 -- 3 param length (4)			*
 *	14,15,16,17 -- 4,5,6,7 value			*
 *							*
 *	18,19 -- 0,1 param code (5)			*
 *	20 -- 2 flags					*
 *	21 -- 3 param length (4)			*
 *	22,23,24,25 -- 4,5,6,7 value			*
 *							*
 *	26,27 -- 0,1 param code (6)			*
 *	28 -- 2 flags					*
 *	29 -- 3 param length (2)			*
 *	30,31 -- 4,5 value				*
 ********************************************************/
static void report_error(
    const char what[],	/* The name of the counter */
    const int page	/* The page number for the report */
)
{
    int total_retries;		/* Total retries corrected (param code=2) */
    int total_corrections;	/* Total errors corrected (3) */
    int total_errors;		/* Total Errors processed (4) */
    long long int total_bytes;	/* Total bytes processed (5) */
    int total_hard;		/* Total hard errors (pc=6) */

    get_page(page);
    if (raw) raw_page();

    total_retries     = THREE_BYTES(&log.raw[8]);
    total_corrections = THREE_BYTES(&log.raw[15]);
    total_errors      = THREE_BYTES(&log.raw[22]);
    total_bytes       = FIVE_BYTES(&log.raw[29]);
    total_hard        = TWO_BYTES(&log.raw[38]);

    if ((total_bytes &       0x8000000000LL) != 0LL) {
	total_bytes |= 0xFFFFFF8000000000LL;
    }


    printf("%s errors: Bytes %5lld Retries %3d Corrections %3d Soft %3d Hard %3d\n",
    	what, 
	total_bytes,
	total_retries, 
	total_corrections,
	total_errors, 
	total_hard);
}
/* ### */
struct counters {
    long int groups_written;
    long int raw_retries;
    long int groups_read;
    long int ecc_retries;
};

/********************************************************
 * usage -- tell the user what to do			*
 ********************************************************/
static void usage(void)
{
    printf("Usage: ex_status [-a] [-e] [-r] device\n");
    printf("Options:\n");
    printf("	-a -- Report all reports\n");
    printf("	-e -- Error report\n");
    printf("	-r -- Print raw block\n");
    exit (8);
}
int main (int argc, char *argv[])
{ 
    /* Report types */
    /* These are go into a bitmapped variable */
    /* and are used to decide which reports to print */

#define R_ALL 	 	-1		/* Report everything */
#define R_ERROR 	(1 << 1)	/* Report error log */

#define R_FIRST 	R_ERROR	/* First report */
#define R_LAST 	 	R_ERROR	/* Second report */

    int report = 0;		/* Reports to print */

    while ((argc > 1) && (argv[1][0] == '-')) {
	switch (argv[1][1]) {
	    case 'e':
		report |= R_ERROR;
		break;
	    case 'a':
		report = R_ALL;
		break;
	    case 'r':
		raw = 1;
		break;
	    default:
		printf("Bad option %s\n", argv[1]);
		usage();
	}
	--argc;
	++argv;
    }
    /* If no reports selected, then use the defaults */
    if (report == 0)
        report = R_ERROR;

    if (argc != 2) {
        usage();
    }

    if ((fd = open(argv[1], O_RDONLY | O_NONBLOCK)) < 0) { 
	if ((fd = open(argv[1], O_RDONLY | O_NONBLOCK)) < 0) { 
	    printf ("Failed to open device %s\n", argv[1]);
	    exit (1);
	}
    }

    if (report == R_ALL)
	report_all();
    else {
	int cur_report;

        for (cur_report = R_FIRST; 
	     cur_report != (R_LAST << 1); 
	     cur_report <<= 1) {

	    if ((cur_report & report) == 0)
	        continue;

	    switch (cur_report) {
		case R_ERROR:
		    report_error("Write", 0x42);
		    report_error("Read ", 0x43);
		    break;
		default:
		    fprintf(stderr, "Internal error: bad report %d\n", 
		    		cur_report);
		    exit (8);
	    }
	}
    }
    return(0);
}
